#include <ATen/core/interned_strings.h>

namespace torch {
namespace jit {

namespace prim {
using namespace ::c10::prim;
}
namespace attr {
using namespace ::c10::attr;
}
namespace aten {
using namespace ::c10::aten;
}
namespace onnx {
using namespace ::c10::onnx;
}
using ::c10::Symbol;
} // namespace jit
} // namespace torch
