#pragma once

// @generated from tools/autograd/templates/python_functions.h

// Python bindings for automatically generated autograd functions

namespace torch { namespace autograd { namespace generated {

void initialize_autogenerated_functions();

}}} // namespace torch::autograd::generated
