## @package onnx
# Module caffe2.python.onnx.error
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
class BaseException(Exception): pass
class Unsupported(BaseException): pass
