#pragma once

// All pure C++ headers for the C++ frontend.
#include <torch/all.h>
// Python bindings for the C++ frontend (includes Python.h).
#include <torch/python.h>
// Deprecated tensor factories (to be removed).
#include <torch/csrc/variable_tensor_functions.h>
