#pragma once

#include <torch/csrc/api/include/torch/types.h>
#include <torch/csrc/autograd/generated/variable_factories.h>
#include <torch/csrc/jit/custom_operator.h>
#include <torch/csrc/jit/import.h>

#include <ATen/ATen.h>
