// Generated by the protocol buffer compiler.  DO NOT EDIT!
// source: caffe2/proto/prof_dag.proto

#ifndef PROTOBUF_caffe2_2fproto_2fprof_5fdag_2eproto__INCLUDED
#define PROTOBUF_caffe2_2fproto_2fprof_5fdag_2eproto__INCLUDED

#include <string>

#include <google/protobuf/stubs/common.h>

#if GOOGLE_PROTOBUF_VERSION < 3005000
#error This file was generated by a newer version of protoc which is
#error incompatible with your Protocol Buffer headers.  Please update
#error your headers.
#endif
#if 3005000 < GOOGLE_PROTOBUF_MIN_PROTOC_VERSION
#error This file was generated by an older version of protoc which is
#error incompatible with your Protocol Buffer headers.  Please
#error regenerate this file with a newer version of protoc.
#endif

#include <google/protobuf/io/coded_stream.h>
#include <google/protobuf/arena.h>
#include <google/protobuf/arenastring.h>
#include <google/protobuf/generated_message_table_driven.h>
#include <google/protobuf/generated_message_util.h>
#include <google/protobuf/metadata.h>
#include <google/protobuf/message.h>
#include <google/protobuf/repeated_field.h>  // IWYU pragma: export
#include <google/protobuf/extension_set.h>  // IWYU pragma: export
#include <google/protobuf/unknown_field_set.h>
// @@protoc_insertion_point(includes)

namespace protobuf_caffe2_2fproto_2fprof_5fdag_2eproto {
// Internal implementation detail -- do not use these members.
struct CAFFE2_API TableStruct {
  static const ::google::protobuf::internal::ParseTableField entries[];
  static const ::google::protobuf::internal::AuxillaryParseTableField aux[];
  static const ::google::protobuf::internal::ParseTable schema[4];
  static const ::google::protobuf::internal::FieldMetadata field_metadata[];
  static const ::google::protobuf::internal::SerializationTable serialization_table[];
  static const ::google::protobuf::uint32 offsets[];
};
void CAFFE2_API AddDescriptors();
void CAFFE2_API InitDefaultsTwoNumberStatsProtoImpl();
void CAFFE2_API InitDefaultsTwoNumberStatsProto();
void CAFFE2_API InitDefaultsBlobProfileImpl();
void CAFFE2_API InitDefaultsBlobProfile();
void CAFFE2_API InitDefaultsProfDAGProtoImpl();
void CAFFE2_API InitDefaultsProfDAGProto();
void CAFFE2_API InitDefaultsProfDAGProtosImpl();
void CAFFE2_API InitDefaultsProfDAGProtos();
inline void CAFFE2_API InitDefaults() {
  InitDefaultsTwoNumberStatsProto();
  InitDefaultsBlobProfile();
  InitDefaultsProfDAGProto();
  InitDefaultsProfDAGProtos();
}
}  // namespace protobuf_caffe2_2fproto_2fprof_5fdag_2eproto
namespace caffe2 { const ::std::string& GetEmptyStringAlreadyInited(); 
class BlobProfile;
class BlobProfileDefaultTypeInternal;
CAFFE2_API extern BlobProfileDefaultTypeInternal _BlobProfile_default_instance_;
class ProfDAGProto;
class ProfDAGProtoDefaultTypeInternal;
CAFFE2_API extern ProfDAGProtoDefaultTypeInternal _ProfDAGProto_default_instance_;
class ProfDAGProtos;
class ProfDAGProtosDefaultTypeInternal;
CAFFE2_API extern ProfDAGProtosDefaultTypeInternal _ProfDAGProtos_default_instance_;
class TwoNumberStatsProto;
class TwoNumberStatsProtoDefaultTypeInternal;
CAFFE2_API extern TwoNumberStatsProtoDefaultTypeInternal _TwoNumberStatsProto_default_instance_;
}  // namespace caffe2
namespace caffe2 {

// ===================================================================

class CAFFE2_API TwoNumberStatsProto : public ::google::protobuf::Message /* @@protoc_insertion_point(class_definition:caffe2.TwoNumberStatsProto) */ {
 public:
  TwoNumberStatsProto();
  virtual ~TwoNumberStatsProto();

  TwoNumberStatsProto(const TwoNumberStatsProto& from);

  inline TwoNumberStatsProto& operator=(const TwoNumberStatsProto& from) {
    CopyFrom(from);
    return *this;
  }
  #if LANG_CXX11
  TwoNumberStatsProto(TwoNumberStatsProto&& from) noexcept
    : TwoNumberStatsProto() {
    *this = ::std::move(from);
  }

  inline TwoNumberStatsProto& operator=(TwoNumberStatsProto&& from) noexcept {
    if (GetArenaNoVirtual() == from.GetArenaNoVirtual()) {
      if (this != &from) InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }
  #endif
  inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const {
    return _internal_metadata_.unknown_fields();
  }
  inline ::google::protobuf::UnknownFieldSet* mutable_unknown_fields() {
    return _internal_metadata_.mutable_unknown_fields();
  }

  static const ::google::protobuf::Descriptor* descriptor();
  static const TwoNumberStatsProto& default_instance();

  static void InitAsDefaultInstance();  // FOR INTERNAL USE ONLY
  static inline const TwoNumberStatsProto* internal_default_instance() {
    return reinterpret_cast<const TwoNumberStatsProto*>(
               &_TwoNumberStatsProto_default_instance_);
  }
  static  int const kIndexInFileMessages =
    0;

  void Swap(TwoNumberStatsProto* other);
  friend void swap(TwoNumberStatsProto& a, TwoNumberStatsProto& b) {
    a.Swap(&b);
  }

  // implements Message ----------------------------------------------

  inline TwoNumberStatsProto* New() const PROTOBUF_FINAL { return New(NULL); }

  TwoNumberStatsProto* New(::google::protobuf::Arena* arena) const PROTOBUF_FINAL;
  void CopyFrom(const ::google::protobuf::Message& from) PROTOBUF_FINAL;
  void MergeFrom(const ::google::protobuf::Message& from) PROTOBUF_FINAL;
  void CopyFrom(const TwoNumberStatsProto& from);
  void MergeFrom(const TwoNumberStatsProto& from);
  void Clear() PROTOBUF_FINAL;
  bool IsInitialized() const PROTOBUF_FINAL;

  size_t ByteSizeLong() const PROTOBUF_FINAL;
  bool MergePartialFromCodedStream(
      ::google::protobuf::io::CodedInputStream* input) PROTOBUF_FINAL;
  void SerializeWithCachedSizes(
      ::google::protobuf::io::CodedOutputStream* output) const PROTOBUF_FINAL;
  ::google::protobuf::uint8* InternalSerializeWithCachedSizesToArray(
      bool deterministic, ::google::protobuf::uint8* target) const PROTOBUF_FINAL;
  int GetCachedSize() const PROTOBUF_FINAL { return _cached_size_; }
  private:
  void SharedCtor();
  void SharedDtor();
  void SetCachedSize(int size) const PROTOBUF_FINAL;
  void InternalSwap(TwoNumberStatsProto* other);
  private:
  inline ::google::protobuf::Arena* GetArenaNoVirtual() const {
    return NULL;
  }
  inline void* MaybeArenaPtr() const {
    return NULL;
  }
  public:

  ::google::protobuf::Metadata GetMetadata() const PROTOBUF_FINAL;

  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------

  // optional float mean = 1;
  bool has_mean() const;
  void clear_mean();
  static const int kMeanFieldNumber = 1;
  float mean() const;
  void set_mean(float value);

  // optional float stddev = 2;
  bool has_stddev() const;
  void clear_stddev();
  static const int kStddevFieldNumber = 2;
  float stddev() const;
  void set_stddev(float value);

  // optional int64 count = 3;
  bool has_count() const;
  void clear_count();
  static const int kCountFieldNumber = 3;
  ::google::protobuf::int64 count() const;
  void set_count(::google::protobuf::int64 value);

  // @@protoc_insertion_point(class_scope:caffe2.TwoNumberStatsProto)
 private:
  void set_has_mean();
  void clear_has_mean();
  void set_has_stddev();
  void clear_has_stddev();
  void set_has_count();
  void clear_has_count();

  ::google::protobuf::internal::InternalMetadataWithArena _internal_metadata_;
  ::google::protobuf::internal::HasBits<1> _has_bits_;
  mutable int _cached_size_;
  float mean_;
  float stddev_;
  ::google::protobuf::int64 count_;
  friend struct ::protobuf_caffe2_2fproto_2fprof_5fdag_2eproto::TableStruct;
  friend void ::protobuf_caffe2_2fproto_2fprof_5fdag_2eproto::InitDefaultsTwoNumberStatsProtoImpl();
};
// -------------------------------------------------------------------

class CAFFE2_API BlobProfile : public ::google::protobuf::Message /* @@protoc_insertion_point(class_definition:caffe2.BlobProfile) */ {
 public:
  BlobProfile();
  virtual ~BlobProfile();

  BlobProfile(const BlobProfile& from);

  inline BlobProfile& operator=(const BlobProfile& from) {
    CopyFrom(from);
    return *this;
  }
  #if LANG_CXX11
  BlobProfile(BlobProfile&& from) noexcept
    : BlobProfile() {
    *this = ::std::move(from);
  }

  inline BlobProfile& operator=(BlobProfile&& from) noexcept {
    if (GetArenaNoVirtual() == from.GetArenaNoVirtual()) {
      if (this != &from) InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }
  #endif
  inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const {
    return _internal_metadata_.unknown_fields();
  }
  inline ::google::protobuf::UnknownFieldSet* mutable_unknown_fields() {
    return _internal_metadata_.mutable_unknown_fields();
  }

  static const ::google::protobuf::Descriptor* descriptor();
  static const BlobProfile& default_instance();

  static void InitAsDefaultInstance();  // FOR INTERNAL USE ONLY
  static inline const BlobProfile* internal_default_instance() {
    return reinterpret_cast<const BlobProfile*>(
               &_BlobProfile_default_instance_);
  }
  static  int const kIndexInFileMessages =
    1;

  void Swap(BlobProfile* other);
  friend void swap(BlobProfile& a, BlobProfile& b) {
    a.Swap(&b);
  }

  // implements Message ----------------------------------------------

  inline BlobProfile* New() const PROTOBUF_FINAL { return New(NULL); }

  BlobProfile* New(::google::protobuf::Arena* arena) const PROTOBUF_FINAL;
  void CopyFrom(const ::google::protobuf::Message& from) PROTOBUF_FINAL;
  void MergeFrom(const ::google::protobuf::Message& from) PROTOBUF_FINAL;
  void CopyFrom(const BlobProfile& from);
  void MergeFrom(const BlobProfile& from);
  void Clear() PROTOBUF_FINAL;
  bool IsInitialized() const PROTOBUF_FINAL;

  size_t ByteSizeLong() const PROTOBUF_FINAL;
  bool MergePartialFromCodedStream(
      ::google::protobuf::io::CodedInputStream* input) PROTOBUF_FINAL;
  void SerializeWithCachedSizes(
      ::google::protobuf::io::CodedOutputStream* output) const PROTOBUF_FINAL;
  ::google::protobuf::uint8* InternalSerializeWithCachedSizesToArray(
      bool deterministic, ::google::protobuf::uint8* target) const PROTOBUF_FINAL;
  int GetCachedSize() const PROTOBUF_FINAL { return _cached_size_; }
  private:
  void SharedCtor();
  void SharedDtor();
  void SetCachedSize(int size) const PROTOBUF_FINAL;
  void InternalSwap(BlobProfile* other);
  private:
  inline ::google::protobuf::Arena* GetArenaNoVirtual() const {
    return NULL;
  }
  inline void* MaybeArenaPtr() const {
    return NULL;
  }
  public:

  ::google::protobuf::Metadata GetMetadata() const PROTOBUF_FINAL;

  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------

  // optional string name = 1;
  bool has_name() const;
  void clear_name();
  static const int kNameFieldNumber = 1;
  const ::std::string& name() const;
  void set_name(const ::std::string& value);
  #if LANG_CXX11
  void set_name(::std::string&& value);
  #endif
  void set_name(const char* value);
  void set_name(const char* value, size_t size);
  ::std::string* mutable_name();
  ::std::string* release_name();
  void set_allocated_name(::std::string* name);

  // optional .caffe2.TwoNumberStatsProto bytes_used = 3;
  bool has_bytes_used() const;
  void clear_bytes_used();
  static const int kBytesUsedFieldNumber = 3;
  const ::caffe2::TwoNumberStatsProto& bytes_used() const;
  ::caffe2::TwoNumberStatsProto* release_bytes_used();
  ::caffe2::TwoNumberStatsProto* mutable_bytes_used();
  void set_allocated_bytes_used(::caffe2::TwoNumberStatsProto* bytes_used);

  // @@protoc_insertion_point(class_scope:caffe2.BlobProfile)
 private:
  void set_has_name();
  void clear_has_name();
  void set_has_bytes_used();
  void clear_has_bytes_used();

  ::google::protobuf::internal::InternalMetadataWithArena _internal_metadata_;
  ::google::protobuf::internal::HasBits<1> _has_bits_;
  mutable int _cached_size_;
  ::google::protobuf::internal::ArenaStringPtr name_;
  ::caffe2::TwoNumberStatsProto* bytes_used_;
  friend struct ::protobuf_caffe2_2fproto_2fprof_5fdag_2eproto::TableStruct;
  friend void ::protobuf_caffe2_2fproto_2fprof_5fdag_2eproto::InitDefaultsBlobProfileImpl();
};
// -------------------------------------------------------------------

class CAFFE2_API ProfDAGProto : public ::google::protobuf::Message /* @@protoc_insertion_point(class_definition:caffe2.ProfDAGProto) */ {
 public:
  ProfDAGProto();
  virtual ~ProfDAGProto();

  ProfDAGProto(const ProfDAGProto& from);

  inline ProfDAGProto& operator=(const ProfDAGProto& from) {
    CopyFrom(from);
    return *this;
  }
  #if LANG_CXX11
  ProfDAGProto(ProfDAGProto&& from) noexcept
    : ProfDAGProto() {
    *this = ::std::move(from);
  }

  inline ProfDAGProto& operator=(ProfDAGProto&& from) noexcept {
    if (GetArenaNoVirtual() == from.GetArenaNoVirtual()) {
      if (this != &from) InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }
  #endif
  inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const {
    return _internal_metadata_.unknown_fields();
  }
  inline ::google::protobuf::UnknownFieldSet* mutable_unknown_fields() {
    return _internal_metadata_.mutable_unknown_fields();
  }

  static const ::google::protobuf::Descriptor* descriptor();
  static const ProfDAGProto& default_instance();

  static void InitAsDefaultInstance();  // FOR INTERNAL USE ONLY
  static inline const ProfDAGProto* internal_default_instance() {
    return reinterpret_cast<const ProfDAGProto*>(
               &_ProfDAGProto_default_instance_);
  }
  static  int const kIndexInFileMessages =
    2;

  void Swap(ProfDAGProto* other);
  friend void swap(ProfDAGProto& a, ProfDAGProto& b) {
    a.Swap(&b);
  }

  // implements Message ----------------------------------------------

  inline ProfDAGProto* New() const PROTOBUF_FINAL { return New(NULL); }

  ProfDAGProto* New(::google::protobuf::Arena* arena) const PROTOBUF_FINAL;
  void CopyFrom(const ::google::protobuf::Message& from) PROTOBUF_FINAL;
  void MergeFrom(const ::google::protobuf::Message& from) PROTOBUF_FINAL;
  void CopyFrom(const ProfDAGProto& from);
  void MergeFrom(const ProfDAGProto& from);
  void Clear() PROTOBUF_FINAL;
  bool IsInitialized() const PROTOBUF_FINAL;

  size_t ByteSizeLong() const PROTOBUF_FINAL;
  bool MergePartialFromCodedStream(
      ::google::protobuf::io::CodedInputStream* input) PROTOBUF_FINAL;
  void SerializeWithCachedSizes(
      ::google::protobuf::io::CodedOutputStream* output) const PROTOBUF_FINAL;
  ::google::protobuf::uint8* InternalSerializeWithCachedSizesToArray(
      bool deterministic, ::google::protobuf::uint8* target) const PROTOBUF_FINAL;
  int GetCachedSize() const PROTOBUF_FINAL { return _cached_size_; }
  private:
  void SharedCtor();
  void SharedDtor();
  void SetCachedSize(int size) const PROTOBUF_FINAL;
  void InternalSwap(ProfDAGProto* other);
  private:
  inline ::google::protobuf::Arena* GetArenaNoVirtual() const {
    return NULL;
  }
  inline void* MaybeArenaPtr() const {
    return NULL;
  }
  public:

  ::google::protobuf::Metadata GetMetadata() const PROTOBUF_FINAL;

  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------

  // repeated .caffe2.BlobProfile output_profile = 5;
  int output_profile_size() const;
  void clear_output_profile();
  static const int kOutputProfileFieldNumber = 5;
  const ::caffe2::BlobProfile& output_profile(int index) const;
  ::caffe2::BlobProfile* mutable_output_profile(int index);
  ::caffe2::BlobProfile* add_output_profile();
  ::google::protobuf::RepeatedPtrField< ::caffe2::BlobProfile >*
      mutable_output_profile();
  const ::google::protobuf::RepeatedPtrField< ::caffe2::BlobProfile >&
      output_profile() const;

  // required string name = 1;
  bool has_name() const;
  void clear_name();
  static const int kNameFieldNumber = 1;
  const ::std::string& name() const;
  void set_name(const ::std::string& value);
  #if LANG_CXX11
  void set_name(::std::string&& value);
  #endif
  void set_name(const char* value);
  void set_name(const char* value, size_t size);
  ::std::string* mutable_name();
  ::std::string* release_name();
  void set_allocated_name(::std::string* name);

  // optional .caffe2.TwoNumberStatsProto execution_time = 4;
  bool has_execution_time() const;
  void clear_execution_time();
  static const int kExecutionTimeFieldNumber = 4;
  const ::caffe2::TwoNumberStatsProto& execution_time() const;
  ::caffe2::TwoNumberStatsProto* release_execution_time();
  ::caffe2::TwoNumberStatsProto* mutable_execution_time();
  void set_allocated_execution_time(::caffe2::TwoNumberStatsProto* execution_time);

  // required float mean = 2;
  bool has_mean() const;
  void clear_mean();
  static const int kMeanFieldNumber = 2;
  float mean() const;
  void set_mean(float value);

  // required float stddev = 3;
  bool has_stddev() const;
  void clear_stddev();
  static const int kStddevFieldNumber = 3;
  float stddev() const;
  void set_stddev(float value);

  // @@protoc_insertion_point(class_scope:caffe2.ProfDAGProto)
 private:
  void set_has_name();
  void clear_has_name();
  void set_has_mean();
  void clear_has_mean();
  void set_has_stddev();
  void clear_has_stddev();
  void set_has_execution_time();
  void clear_has_execution_time();

  // helper for ByteSizeLong()
  size_t RequiredFieldsByteSizeFallback() const;

  ::google::protobuf::internal::InternalMetadataWithArena _internal_metadata_;
  ::google::protobuf::internal::HasBits<1> _has_bits_;
  mutable int _cached_size_;
  ::google::protobuf::RepeatedPtrField< ::caffe2::BlobProfile > output_profile_;
  ::google::protobuf::internal::ArenaStringPtr name_;
  ::caffe2::TwoNumberStatsProto* execution_time_;
  float mean_;
  float stddev_;
  friend struct ::protobuf_caffe2_2fproto_2fprof_5fdag_2eproto::TableStruct;
  friend void ::protobuf_caffe2_2fproto_2fprof_5fdag_2eproto::InitDefaultsProfDAGProtoImpl();
};
// -------------------------------------------------------------------

class CAFFE2_API ProfDAGProtos : public ::google::protobuf::Message /* @@protoc_insertion_point(class_definition:caffe2.ProfDAGProtos) */ {
 public:
  ProfDAGProtos();
  virtual ~ProfDAGProtos();

  ProfDAGProtos(const ProfDAGProtos& from);

  inline ProfDAGProtos& operator=(const ProfDAGProtos& from) {
    CopyFrom(from);
    return *this;
  }
  #if LANG_CXX11
  ProfDAGProtos(ProfDAGProtos&& from) noexcept
    : ProfDAGProtos() {
    *this = ::std::move(from);
  }

  inline ProfDAGProtos& operator=(ProfDAGProtos&& from) noexcept {
    if (GetArenaNoVirtual() == from.GetArenaNoVirtual()) {
      if (this != &from) InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }
  #endif
  inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const {
    return _internal_metadata_.unknown_fields();
  }
  inline ::google::protobuf::UnknownFieldSet* mutable_unknown_fields() {
    return _internal_metadata_.mutable_unknown_fields();
  }

  static const ::google::protobuf::Descriptor* descriptor();
  static const ProfDAGProtos& default_instance();

  static void InitAsDefaultInstance();  // FOR INTERNAL USE ONLY
  static inline const ProfDAGProtos* internal_default_instance() {
    return reinterpret_cast<const ProfDAGProtos*>(
               &_ProfDAGProtos_default_instance_);
  }
  static  int const kIndexInFileMessages =
    3;

  void Swap(ProfDAGProtos* other);
  friend void swap(ProfDAGProtos& a, ProfDAGProtos& b) {
    a.Swap(&b);
  }

  // implements Message ----------------------------------------------

  inline ProfDAGProtos* New() const PROTOBUF_FINAL { return New(NULL); }

  ProfDAGProtos* New(::google::protobuf::Arena* arena) const PROTOBUF_FINAL;
  void CopyFrom(const ::google::protobuf::Message& from) PROTOBUF_FINAL;
  void MergeFrom(const ::google::protobuf::Message& from) PROTOBUF_FINAL;
  void CopyFrom(const ProfDAGProtos& from);
  void MergeFrom(const ProfDAGProtos& from);
  void Clear() PROTOBUF_FINAL;
  bool IsInitialized() const PROTOBUF_FINAL;

  size_t ByteSizeLong() const PROTOBUF_FINAL;
  bool MergePartialFromCodedStream(
      ::google::protobuf::io::CodedInputStream* input) PROTOBUF_FINAL;
  void SerializeWithCachedSizes(
      ::google::protobuf::io::CodedOutputStream* output) const PROTOBUF_FINAL;
  ::google::protobuf::uint8* InternalSerializeWithCachedSizesToArray(
      bool deterministic, ::google::protobuf::uint8* target) const PROTOBUF_FINAL;
  int GetCachedSize() const PROTOBUF_FINAL { return _cached_size_; }
  private:
  void SharedCtor();
  void SharedDtor();
  void SetCachedSize(int size) const PROTOBUF_FINAL;
  void InternalSwap(ProfDAGProtos* other);
  private:
  inline ::google::protobuf::Arena* GetArenaNoVirtual() const {
    return NULL;
  }
  inline void* MaybeArenaPtr() const {
    return NULL;
  }
  public:

  ::google::protobuf::Metadata GetMetadata() const PROTOBUF_FINAL;

  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------

  // repeated .caffe2.ProfDAGProto stats = 1;
  int stats_size() const;
  void clear_stats();
  static const int kStatsFieldNumber = 1;
  const ::caffe2::ProfDAGProto& stats(int index) const;
  ::caffe2::ProfDAGProto* mutable_stats(int index);
  ::caffe2::ProfDAGProto* add_stats();
  ::google::protobuf::RepeatedPtrField< ::caffe2::ProfDAGProto >*
      mutable_stats();
  const ::google::protobuf::RepeatedPtrField< ::caffe2::ProfDAGProto >&
      stats() const;

  // optional string net_name = 2;
  bool has_net_name() const;
  void clear_net_name();
  static const int kNetNameFieldNumber = 2;
  const ::std::string& net_name() const;
  void set_net_name(const ::std::string& value);
  #if LANG_CXX11
  void set_net_name(::std::string&& value);
  #endif
  void set_net_name(const char* value);
  void set_net_name(const char* value, size_t size);
  ::std::string* mutable_net_name();
  ::std::string* release_net_name();
  void set_allocated_net_name(::std::string* net_name);

  // @@protoc_insertion_point(class_scope:caffe2.ProfDAGProtos)
 private:
  void set_has_net_name();
  void clear_has_net_name();

  ::google::protobuf::internal::InternalMetadataWithArena _internal_metadata_;
  ::google::protobuf::internal::HasBits<1> _has_bits_;
  mutable int _cached_size_;
  ::google::protobuf::RepeatedPtrField< ::caffe2::ProfDAGProto > stats_;
  ::google::protobuf::internal::ArenaStringPtr net_name_;
  friend struct ::protobuf_caffe2_2fproto_2fprof_5fdag_2eproto::TableStruct;
  friend void ::protobuf_caffe2_2fproto_2fprof_5fdag_2eproto::InitDefaultsProfDAGProtosImpl();
};
// ===================================================================


// ===================================================================

#ifdef __GNUC__
  #pragma GCC diagnostic push
  #pragma GCC diagnostic ignored "-Wstrict-aliasing"
#endif  // __GNUC__
// TwoNumberStatsProto

// optional float mean = 1;
inline bool TwoNumberStatsProto::has_mean() const {
  return (_has_bits_[0] & 0x00000001u) != 0;
}
inline void TwoNumberStatsProto::set_has_mean() {
  _has_bits_[0] |= 0x00000001u;
}
inline void TwoNumberStatsProto::clear_has_mean() {
  _has_bits_[0] &= ~0x00000001u;
}
inline void TwoNumberStatsProto::clear_mean() {
  mean_ = 0;
  clear_has_mean();
}
inline float TwoNumberStatsProto::mean() const {
  // @@protoc_insertion_point(field_get:caffe2.TwoNumberStatsProto.mean)
  return mean_;
}
inline void TwoNumberStatsProto::set_mean(float value) {
  set_has_mean();
  mean_ = value;
  // @@protoc_insertion_point(field_set:caffe2.TwoNumberStatsProto.mean)
}

// optional float stddev = 2;
inline bool TwoNumberStatsProto::has_stddev() const {
  return (_has_bits_[0] & 0x00000002u) != 0;
}
inline void TwoNumberStatsProto::set_has_stddev() {
  _has_bits_[0] |= 0x00000002u;
}
inline void TwoNumberStatsProto::clear_has_stddev() {
  _has_bits_[0] &= ~0x00000002u;
}
inline void TwoNumberStatsProto::clear_stddev() {
  stddev_ = 0;
  clear_has_stddev();
}
inline float TwoNumberStatsProto::stddev() const {
  // @@protoc_insertion_point(field_get:caffe2.TwoNumberStatsProto.stddev)
  return stddev_;
}
inline void TwoNumberStatsProto::set_stddev(float value) {
  set_has_stddev();
  stddev_ = value;
  // @@protoc_insertion_point(field_set:caffe2.TwoNumberStatsProto.stddev)
}

// optional int64 count = 3;
inline bool TwoNumberStatsProto::has_count() const {
  return (_has_bits_[0] & 0x00000004u) != 0;
}
inline void TwoNumberStatsProto::set_has_count() {
  _has_bits_[0] |= 0x00000004u;
}
inline void TwoNumberStatsProto::clear_has_count() {
  _has_bits_[0] &= ~0x00000004u;
}
inline void TwoNumberStatsProto::clear_count() {
  count_ = GOOGLE_LONGLONG(0);
  clear_has_count();
}
inline ::google::protobuf::int64 TwoNumberStatsProto::count() const {
  // @@protoc_insertion_point(field_get:caffe2.TwoNumberStatsProto.count)
  return count_;
}
inline void TwoNumberStatsProto::set_count(::google::protobuf::int64 value) {
  set_has_count();
  count_ = value;
  // @@protoc_insertion_point(field_set:caffe2.TwoNumberStatsProto.count)
}

// -------------------------------------------------------------------

// BlobProfile

// optional string name = 1;
inline bool BlobProfile::has_name() const {
  return (_has_bits_[0] & 0x00000001u) != 0;
}
inline void BlobProfile::set_has_name() {
  _has_bits_[0] |= 0x00000001u;
}
inline void BlobProfile::clear_has_name() {
  _has_bits_[0] &= ~0x00000001u;
}
inline void BlobProfile::clear_name() {
  name_.ClearToEmptyNoArena(&GetEmptyStringAlreadyInited());
  clear_has_name();
}
inline const ::std::string& BlobProfile::name() const {
  // @@protoc_insertion_point(field_get:caffe2.BlobProfile.name)
  return name_.GetNoArena();
}
inline void BlobProfile::set_name(const ::std::string& value) {
  set_has_name();
  name_.SetNoArena(&GetEmptyStringAlreadyInited(), value);
  // @@protoc_insertion_point(field_set:caffe2.BlobProfile.name)
}
#if LANG_CXX11
inline void BlobProfile::set_name(::std::string&& value) {
  set_has_name();
  name_.SetNoArena(
    &GetEmptyStringAlreadyInited(), ::std::move(value));
  // @@protoc_insertion_point(field_set_rvalue:caffe2.BlobProfile.name)
}
#endif
inline void BlobProfile::set_name(const char* value) {
  GOOGLE_DCHECK(value != NULL);
  set_has_name();
  name_.SetNoArena(&GetEmptyStringAlreadyInited(), ::std::string(value));
  // @@protoc_insertion_point(field_set_char:caffe2.BlobProfile.name)
}
inline void BlobProfile::set_name(const char* value, size_t size) {
  set_has_name();
  name_.SetNoArena(&GetEmptyStringAlreadyInited(),
      ::std::string(reinterpret_cast<const char*>(value), size));
  // @@protoc_insertion_point(field_set_pointer:caffe2.BlobProfile.name)
}
inline ::std::string* BlobProfile::mutable_name() {
  set_has_name();
  // @@protoc_insertion_point(field_mutable:caffe2.BlobProfile.name)
  return name_.MutableNoArena(&GetEmptyStringAlreadyInited());
}
inline ::std::string* BlobProfile::release_name() {
  // @@protoc_insertion_point(field_release:caffe2.BlobProfile.name)
  clear_has_name();
  return name_.ReleaseNoArena(&GetEmptyStringAlreadyInited());
}
inline void BlobProfile::set_allocated_name(::std::string* name) {
  if (name != NULL) {
    set_has_name();
  } else {
    clear_has_name();
  }
  name_.SetAllocatedNoArena(&GetEmptyStringAlreadyInited(), name);
  // @@protoc_insertion_point(field_set_allocated:caffe2.BlobProfile.name)
}

// optional .caffe2.TwoNumberStatsProto bytes_used = 3;
inline bool BlobProfile::has_bytes_used() const {
  return (_has_bits_[0] & 0x00000002u) != 0;
}
inline void BlobProfile::set_has_bytes_used() {
  _has_bits_[0] |= 0x00000002u;
}
inline void BlobProfile::clear_has_bytes_used() {
  _has_bits_[0] &= ~0x00000002u;
}
inline void BlobProfile::clear_bytes_used() {
  if (bytes_used_ != NULL) bytes_used_->Clear();
  clear_has_bytes_used();
}
inline const ::caffe2::TwoNumberStatsProto& BlobProfile::bytes_used() const {
  const ::caffe2::TwoNumberStatsProto* p = bytes_used_;
  // @@protoc_insertion_point(field_get:caffe2.BlobProfile.bytes_used)
  return p != NULL ? *p : *reinterpret_cast<const ::caffe2::TwoNumberStatsProto*>(
      &::caffe2::_TwoNumberStatsProto_default_instance_);
}
inline ::caffe2::TwoNumberStatsProto* BlobProfile::release_bytes_used() {
  // @@protoc_insertion_point(field_release:caffe2.BlobProfile.bytes_used)
  clear_has_bytes_used();
  ::caffe2::TwoNumberStatsProto* temp = bytes_used_;
  bytes_used_ = NULL;
  return temp;
}
inline ::caffe2::TwoNumberStatsProto* BlobProfile::mutable_bytes_used() {
  set_has_bytes_used();
  if (bytes_used_ == NULL) {
    bytes_used_ = new ::caffe2::TwoNumberStatsProto;
  }
  // @@protoc_insertion_point(field_mutable:caffe2.BlobProfile.bytes_used)
  return bytes_used_;
}
inline void BlobProfile::set_allocated_bytes_used(::caffe2::TwoNumberStatsProto* bytes_used) {
  ::google::protobuf::Arena* message_arena = GetArenaNoVirtual();
  if (message_arena == NULL) {
    delete bytes_used_;
  }
  if (bytes_used) {
    ::google::protobuf::Arena* submessage_arena = NULL;
    if (message_arena != submessage_arena) {
      bytes_used = ::google::protobuf::internal::GetOwnedMessage(
          message_arena, bytes_used, submessage_arena);
    }
    set_has_bytes_used();
  } else {
    clear_has_bytes_used();
  }
  bytes_used_ = bytes_used;
  // @@protoc_insertion_point(field_set_allocated:caffe2.BlobProfile.bytes_used)
}

// -------------------------------------------------------------------

// ProfDAGProto

// required string name = 1;
inline bool ProfDAGProto::has_name() const {
  return (_has_bits_[0] & 0x00000001u) != 0;
}
inline void ProfDAGProto::set_has_name() {
  _has_bits_[0] |= 0x00000001u;
}
inline void ProfDAGProto::clear_has_name() {
  _has_bits_[0] &= ~0x00000001u;
}
inline void ProfDAGProto::clear_name() {
  name_.ClearToEmptyNoArena(&GetEmptyStringAlreadyInited());
  clear_has_name();
}
inline const ::std::string& ProfDAGProto::name() const {
  // @@protoc_insertion_point(field_get:caffe2.ProfDAGProto.name)
  return name_.GetNoArena();
}
inline void ProfDAGProto::set_name(const ::std::string& value) {
  set_has_name();
  name_.SetNoArena(&GetEmptyStringAlreadyInited(), value);
  // @@protoc_insertion_point(field_set:caffe2.ProfDAGProto.name)
}
#if LANG_CXX11
inline void ProfDAGProto::set_name(::std::string&& value) {
  set_has_name();
  name_.SetNoArena(
    &GetEmptyStringAlreadyInited(), ::std::move(value));
  // @@protoc_insertion_point(field_set_rvalue:caffe2.ProfDAGProto.name)
}
#endif
inline void ProfDAGProto::set_name(const char* value) {
  GOOGLE_DCHECK(value != NULL);
  set_has_name();
  name_.SetNoArena(&GetEmptyStringAlreadyInited(), ::std::string(value));
  // @@protoc_insertion_point(field_set_char:caffe2.ProfDAGProto.name)
}
inline void ProfDAGProto::set_name(const char* value, size_t size) {
  set_has_name();
  name_.SetNoArena(&GetEmptyStringAlreadyInited(),
      ::std::string(reinterpret_cast<const char*>(value), size));
  // @@protoc_insertion_point(field_set_pointer:caffe2.ProfDAGProto.name)
}
inline ::std::string* ProfDAGProto::mutable_name() {
  set_has_name();
  // @@protoc_insertion_point(field_mutable:caffe2.ProfDAGProto.name)
  return name_.MutableNoArena(&GetEmptyStringAlreadyInited());
}
inline ::std::string* ProfDAGProto::release_name() {
  // @@protoc_insertion_point(field_release:caffe2.ProfDAGProto.name)
  clear_has_name();
  return name_.ReleaseNoArena(&GetEmptyStringAlreadyInited());
}
inline void ProfDAGProto::set_allocated_name(::std::string* name) {
  if (name != NULL) {
    set_has_name();
  } else {
    clear_has_name();
  }
  name_.SetAllocatedNoArena(&GetEmptyStringAlreadyInited(), name);
  // @@protoc_insertion_point(field_set_allocated:caffe2.ProfDAGProto.name)
}

// required float mean = 2;
inline bool ProfDAGProto::has_mean() const {
  return (_has_bits_[0] & 0x00000004u) != 0;
}
inline void ProfDAGProto::set_has_mean() {
  _has_bits_[0] |= 0x00000004u;
}
inline void ProfDAGProto::clear_has_mean() {
  _has_bits_[0] &= ~0x00000004u;
}
inline void ProfDAGProto::clear_mean() {
  mean_ = 0;
  clear_has_mean();
}
inline float ProfDAGProto::mean() const {
  // @@protoc_insertion_point(field_get:caffe2.ProfDAGProto.mean)
  return mean_;
}
inline void ProfDAGProto::set_mean(float value) {
  set_has_mean();
  mean_ = value;
  // @@protoc_insertion_point(field_set:caffe2.ProfDAGProto.mean)
}

// required float stddev = 3;
inline bool ProfDAGProto::has_stddev() const {
  return (_has_bits_[0] & 0x00000008u) != 0;
}
inline void ProfDAGProto::set_has_stddev() {
  _has_bits_[0] |= 0x00000008u;
}
inline void ProfDAGProto::clear_has_stddev() {
  _has_bits_[0] &= ~0x00000008u;
}
inline void ProfDAGProto::clear_stddev() {
  stddev_ = 0;
  clear_has_stddev();
}
inline float ProfDAGProto::stddev() const {
  // @@protoc_insertion_point(field_get:caffe2.ProfDAGProto.stddev)
  return stddev_;
}
inline void ProfDAGProto::set_stddev(float value) {
  set_has_stddev();
  stddev_ = value;
  // @@protoc_insertion_point(field_set:caffe2.ProfDAGProto.stddev)
}

// optional .caffe2.TwoNumberStatsProto execution_time = 4;
inline bool ProfDAGProto::has_execution_time() const {
  return (_has_bits_[0] & 0x00000002u) != 0;
}
inline void ProfDAGProto::set_has_execution_time() {
  _has_bits_[0] |= 0x00000002u;
}
inline void ProfDAGProto::clear_has_execution_time() {
  _has_bits_[0] &= ~0x00000002u;
}
inline void ProfDAGProto::clear_execution_time() {
  if (execution_time_ != NULL) execution_time_->Clear();
  clear_has_execution_time();
}
inline const ::caffe2::TwoNumberStatsProto& ProfDAGProto::execution_time() const {
  const ::caffe2::TwoNumberStatsProto* p = execution_time_;
  // @@protoc_insertion_point(field_get:caffe2.ProfDAGProto.execution_time)
  return p != NULL ? *p : *reinterpret_cast<const ::caffe2::TwoNumberStatsProto*>(
      &::caffe2::_TwoNumberStatsProto_default_instance_);
}
inline ::caffe2::TwoNumberStatsProto* ProfDAGProto::release_execution_time() {
  // @@protoc_insertion_point(field_release:caffe2.ProfDAGProto.execution_time)
  clear_has_execution_time();
  ::caffe2::TwoNumberStatsProto* temp = execution_time_;
  execution_time_ = NULL;
  return temp;
}
inline ::caffe2::TwoNumberStatsProto* ProfDAGProto::mutable_execution_time() {
  set_has_execution_time();
  if (execution_time_ == NULL) {
    execution_time_ = new ::caffe2::TwoNumberStatsProto;
  }
  // @@protoc_insertion_point(field_mutable:caffe2.ProfDAGProto.execution_time)
  return execution_time_;
}
inline void ProfDAGProto::set_allocated_execution_time(::caffe2::TwoNumberStatsProto* execution_time) {
  ::google::protobuf::Arena* message_arena = GetArenaNoVirtual();
  if (message_arena == NULL) {
    delete execution_time_;
  }
  if (execution_time) {
    ::google::protobuf::Arena* submessage_arena = NULL;
    if (message_arena != submessage_arena) {
      execution_time = ::google::protobuf::internal::GetOwnedMessage(
          message_arena, execution_time, submessage_arena);
    }
    set_has_execution_time();
  } else {
    clear_has_execution_time();
  }
  execution_time_ = execution_time;
  // @@protoc_insertion_point(field_set_allocated:caffe2.ProfDAGProto.execution_time)
}

// repeated .caffe2.BlobProfile output_profile = 5;
inline int ProfDAGProto::output_profile_size() const {
  return output_profile_.size();
}
inline void ProfDAGProto::clear_output_profile() {
  output_profile_.Clear();
}
inline const ::caffe2::BlobProfile& ProfDAGProto::output_profile(int index) const {
  // @@protoc_insertion_point(field_get:caffe2.ProfDAGProto.output_profile)
  return output_profile_.Get(index);
}
inline ::caffe2::BlobProfile* ProfDAGProto::mutable_output_profile(int index) {
  // @@protoc_insertion_point(field_mutable:caffe2.ProfDAGProto.output_profile)
  return output_profile_.Mutable(index);
}
inline ::caffe2::BlobProfile* ProfDAGProto::add_output_profile() {
  // @@protoc_insertion_point(field_add:caffe2.ProfDAGProto.output_profile)
  return output_profile_.Add();
}
inline ::google::protobuf::RepeatedPtrField< ::caffe2::BlobProfile >*
ProfDAGProto::mutable_output_profile() {
  // @@protoc_insertion_point(field_mutable_list:caffe2.ProfDAGProto.output_profile)
  return &output_profile_;
}
inline const ::google::protobuf::RepeatedPtrField< ::caffe2::BlobProfile >&
ProfDAGProto::output_profile() const {
  // @@protoc_insertion_point(field_list:caffe2.ProfDAGProto.output_profile)
  return output_profile_;
}

// -------------------------------------------------------------------

// ProfDAGProtos

// repeated .caffe2.ProfDAGProto stats = 1;
inline int ProfDAGProtos::stats_size() const {
  return stats_.size();
}
inline void ProfDAGProtos::clear_stats() {
  stats_.Clear();
}
inline const ::caffe2::ProfDAGProto& ProfDAGProtos::stats(int index) const {
  // @@protoc_insertion_point(field_get:caffe2.ProfDAGProtos.stats)
  return stats_.Get(index);
}
inline ::caffe2::ProfDAGProto* ProfDAGProtos::mutable_stats(int index) {
  // @@protoc_insertion_point(field_mutable:caffe2.ProfDAGProtos.stats)
  return stats_.Mutable(index);
}
inline ::caffe2::ProfDAGProto* ProfDAGProtos::add_stats() {
  // @@protoc_insertion_point(field_add:caffe2.ProfDAGProtos.stats)
  return stats_.Add();
}
inline ::google::protobuf::RepeatedPtrField< ::caffe2::ProfDAGProto >*
ProfDAGProtos::mutable_stats() {
  // @@protoc_insertion_point(field_mutable_list:caffe2.ProfDAGProtos.stats)
  return &stats_;
}
inline const ::google::protobuf::RepeatedPtrField< ::caffe2::ProfDAGProto >&
ProfDAGProtos::stats() const {
  // @@protoc_insertion_point(field_list:caffe2.ProfDAGProtos.stats)
  return stats_;
}

// optional string net_name = 2;
inline bool ProfDAGProtos::has_net_name() const {
  return (_has_bits_[0] & 0x00000001u) != 0;
}
inline void ProfDAGProtos::set_has_net_name() {
  _has_bits_[0] |= 0x00000001u;
}
inline void ProfDAGProtos::clear_has_net_name() {
  _has_bits_[0] &= ~0x00000001u;
}
inline void ProfDAGProtos::clear_net_name() {
  net_name_.ClearToEmptyNoArena(&GetEmptyStringAlreadyInited());
  clear_has_net_name();
}
inline const ::std::string& ProfDAGProtos::net_name() const {
  // @@protoc_insertion_point(field_get:caffe2.ProfDAGProtos.net_name)
  return net_name_.GetNoArena();
}
inline void ProfDAGProtos::set_net_name(const ::std::string& value) {
  set_has_net_name();
  net_name_.SetNoArena(&GetEmptyStringAlreadyInited(), value);
  // @@protoc_insertion_point(field_set:caffe2.ProfDAGProtos.net_name)
}
#if LANG_CXX11
inline void ProfDAGProtos::set_net_name(::std::string&& value) {
  set_has_net_name();
  net_name_.SetNoArena(
    &GetEmptyStringAlreadyInited(), ::std::move(value));
  // @@protoc_insertion_point(field_set_rvalue:caffe2.ProfDAGProtos.net_name)
}
#endif
inline void ProfDAGProtos::set_net_name(const char* value) {
  GOOGLE_DCHECK(value != NULL);
  set_has_net_name();
  net_name_.SetNoArena(&GetEmptyStringAlreadyInited(), ::std::string(value));
  // @@protoc_insertion_point(field_set_char:caffe2.ProfDAGProtos.net_name)
}
inline void ProfDAGProtos::set_net_name(const char* value, size_t size) {
  set_has_net_name();
  net_name_.SetNoArena(&GetEmptyStringAlreadyInited(),
      ::std::string(reinterpret_cast<const char*>(value), size));
  // @@protoc_insertion_point(field_set_pointer:caffe2.ProfDAGProtos.net_name)
}
inline ::std::string* ProfDAGProtos::mutable_net_name() {
  set_has_net_name();
  // @@protoc_insertion_point(field_mutable:caffe2.ProfDAGProtos.net_name)
  return net_name_.MutableNoArena(&GetEmptyStringAlreadyInited());
}
inline ::std::string* ProfDAGProtos::release_net_name() {
  // @@protoc_insertion_point(field_release:caffe2.ProfDAGProtos.net_name)
  clear_has_net_name();
  return net_name_.ReleaseNoArena(&GetEmptyStringAlreadyInited());
}
inline void ProfDAGProtos::set_allocated_net_name(::std::string* net_name) {
  if (net_name != NULL) {
    set_has_net_name();
  } else {
    clear_has_net_name();
  }
  net_name_.SetAllocatedNoArena(&GetEmptyStringAlreadyInited(), net_name);
  // @@protoc_insertion_point(field_set_allocated:caffe2.ProfDAGProtos.net_name)
}

#ifdef __GNUC__
  #pragma GCC diagnostic pop
#endif  // __GNUC__
// -------------------------------------------------------------------

// -------------------------------------------------------------------

// -------------------------------------------------------------------


// @@protoc_insertion_point(namespace_scope)

}  // namespace caffe2

// @@protoc_insertion_point(global_scope)

#endif  // PROTOBUF_caffe2_2fproto_2fprof_5fdag_2eproto__INCLUDED
